package main

import (
	"github.com/jessevdk/go-flags"

	"github.com/canonical/chisel/internal/archive"
	"github.com/canonical/chisel/internal/cache"
	"github.com/canonical/chisel/internal/setup"
	"github.com/canonical/chisel/internal/slicer"
)

var shortCutHelp = "Cut a tree with selected slices"
var longCutHelp = `
The cut command uses the provided selection of package slices
to create a new filesystem tree in the root location.

By default it fetches the slices for the same OpenCloudOS/TencentOS version as the
current host, unless the --release flag is used.
`

var cutDescs = map[string]string{
	"release": "Chisel release name or directory (e.g. opencloudos-stream-23)",
	"root":    "Root for generated content",
	"arch":    "Package architecture",
}

type cmdCut struct {
	Release string `long:"release" value-name:"<dir>"`
	RootDir string `long:"root" value-name:"<dir>" required:"yes"`
	Arch    string `long:"arch" value-name:"<arch>"`

	Positional struct {
		SliceRefs []string `positional-arg-name:"<slice names>" required:"yes"`
	} `positional-args:"yes"`
}

func init() {
	addCommand("cut", shortCutHelp, longCutHelp, func() flags.Commander { return &cmdCut{} }, cutDescs, nil)
}

func (cmd *cmdCut) Execute(args []string) error {
	if len(args) > 0 {
		return ErrExtraArgs
	}

	sliceKeys := make([]setup.SliceKey, len(cmd.Positional.SliceRefs))
	for i, sliceRef := range cmd.Positional.SliceRefs {
		sliceKey, err := setup.ParseSliceKey(sliceRef)
		if err != nil {
			return err
		}
		sliceKeys[i] = sliceKey
	}
	logf("===========================一、执行解析chisel-release仓库%s分支的内容==========================", cmd.Release)
	// 读取解析chisel-release仓库具体内容，保存在release中
	release, err := obtainRelease(cmd.Release)
	if err != nil {
		return err
	}
	logf("===========================二、获取切片对应的依赖切片===========================")
	// 从release中获取sliceKeys所有的依赖切片，这里面release没动过
	selection, err := setup.Select(release, sliceKeys)
	if err != nil {
		return err
	}
	logf("执行完Select后返回的数据selection: \n %+v", selection)

	//这里archive不是setup中的archive结构体，而是archive.Archive接口定义
	archives := make(map[string]archive.Archive)
	// release.Archives是setup里的Archive的结构体
	logf("===========================三、获取repodata里所有rpm包的相关数据==========================")
	//一般来说，一个release里只有一个archive, 例如：https://github.com/wynnfeng/chisel-releases/blob/opencloudos-stream-23/chisel.yaml
	for archiveName, archiveInfo := range release.Archives {
		logf("ArchiveName: %q", archiveName)
		openArchive, err := archive.Open(&archive.Options{
			Label:      archiveName,
			Version:    archiveInfo.Version,
			Arch:       cmd.Arch,
			Suites:     archiveInfo.Suites,
			Components: archiveInfo.Components,
			Pro:        archiveInfo.Pro,
			CacheDir:   cache.DefaultDir("chisel"),
			PubKeys:    archiveInfo.PubKeys,
		})
		if err != nil {
			if err == archive.ErrCredentialsNotFound {
				logf("Archive %q ignored: credentials not found", archiveName)
				continue
			}
			return err
		}
		//openArchive解析并保存了repodata里每个xml数据在indexs里，参考tencentosArchive的定义，除此之外它还是Archive接口的实现
		//openArchive也包含了release.Archives里的数据，保存在option结构体里
		//可以理解为openArchive不仅包含了release.Archives里的数据，它同时获取到repodata对应suites、components下的xml数据，这个数据记录了rpm包相关信息
		archives[archiveName] = openArchive
	}
	logf("===========================四、下载安装对应的rpm包, 根据切片提取文件==========================")
	err = slicer.Run(&slicer.RunOptions{
		Selection: selection,
		Archives:  archives,
		TargetDir: cmd.RootDir,
	})
	logf("切片完成")
	return err
}
