package main

import (
	"fmt"
	"io"
	"log"
	"os"
	"strings"
	"unicode"
	"unicode/utf8"

	"github.com/jessevdk/go-flags"
	"golang.org/x/term"

	"github.com/canonical/chisel/internal/archive"
	"github.com/canonical/chisel/internal/rpm"
	"github.com/canonical/chisel/internal/setup"
	"github.com/canonical/chisel/internal/slicer"
)

var (
	// Standard streams, redirected for testing.
	Stdin  io.Reader = os.Stdin
	Stdout io.Writer = os.Stdout
	Stderr io.Writer = os.Stderr
	// overridden for testing
	ReadPassword = term.ReadPassword
	// set to logger.Panicf in testing
	//noticef = logger.Noticef
)

type options struct {
	Version func() `long:"version"`
}

type argDesc struct {
	name string
	desc string
}

var optionsData options

// ErrExtraArgs is returned  if extra arguments to a command are found
var ErrExtraArgs = fmt.Errorf("too many arguments for command")

// cmdInfo holds information needed to call parser.AddCommand(...).
type cmdInfo struct {
	name, shortHelp, longHelp string
	builder                   func() flags.Commander
	hidden                    bool
	optDescs                  map[string]string
	argDescs                  []argDesc
	alias                     string
	extra                     func(*flags.Command)
}

// commands holds information about all non-debug commands.
var commands []*cmdInfo

// debugCommands holds information about all debug commands.
var debugCommands []*cmdInfo

// addCommand replaces parser.addCommand() in a way that is compatible with
// re-constructing a pristine parser.
func addCommand(name, shortHelp, longHelp string, builder func() flags.Commander, optDescs map[string]string, argDescs []argDesc) *cmdInfo {
	info := &cmdInfo{
		name:      name,
		shortHelp: shortHelp,
		longHelp:  longHelp,
		builder:   builder,
		optDescs:  optDescs,
		argDescs:  argDescs,
	}
	commands = append(commands, info)
	return info
}

// addDebugCommand replaces parser.addCommand() in a way that is
// compatible with re-constructing a pristine parser. It is meant for
// adding debug commands.
func addDebugCommand(name, shortHelp, longHelp string, builder func() flags.Commander, optDescs map[string]string, argDescs []argDesc) *cmdInfo {
	info := &cmdInfo{
		name:      name,
		shortHelp: shortHelp,
		longHelp:  longHelp,
		builder:   builder,
		optDescs:  optDescs,
		argDescs:  argDescs,
	}
	debugCommands = append(debugCommands, info)
	return info
}

type parserSetter interface {
	setParser(*flags.Parser)
}

func lintDesc(cmdName, optName, desc, origDesc string) {
	if len(optName) == 0 {
		panicf("option on %q has no name", cmdName)
	}
	if len(origDesc) != 0 {
		panicf("description of %s's %q of %q set from tag", cmdName, optName, origDesc)
	}
	if len(desc) > 0 {
		// decode the first rune instead of converting all of desc into []rune
		r, _ := utf8.DecodeRuneInString(desc)
		// note IsLower != !IsUpper for runes with no upper/lower.
		if unicode.IsLower(r) && !strings.HasPrefix(desc, "login.ubuntu.com") && !strings.HasPrefix(desc, cmdName) {
			logf("description of %s's %q is lowercase: %q", cmdName, optName, desc)
		}
	}
}

func lintArg(cmdName, optName, desc, origDesc string) {
	lintDesc(cmdName, optName, desc, origDesc)
	if len(optName) > 0 && optName[0] == '<' && optName[len(optName)-1] == '>' {
		return
	}
	if len(optName) > 0 && optName[0] == '<' && strings.HasSuffix(optName, ">s") {
		// see comment in fixupArg about the >s case
		return
	}
	logf("argument %q's %q should begin with < and end with >", cmdName, optName)
}

func fixupArg(optName string) string {
	// Due to misunderstanding some localized versions of option name are
	// literally "<option>s" instead of "<option>". While translators can
	// improve this over time we can be smarter and avoid silly messages
	// logged whenever the command is used.
	//
	// See: https://bugs.launchpad.net/snapd/+bug/1806761
	if strings.HasSuffix(optName, ">s") {
		return optName[:len(optName)-1]
	}
	return optName
}

//type clientSetter interface {
//	setClient(*client.Client)
//}
//
//type clientMixin struct {
//	client *client.Client
//}
//
//func (ch *clientMixin) setClient(cli *client.Client) {
//	ch.client = cli
//}

// Parser creates and populates a fresh parser.
// Since commands have local state a fresh parser is required to isolate tests
// from each other.
func Parser() *flags.Parser {
	// 1.optionsData是options结构体，当用户运行程序时指定 --version 参数，
	// 2.go-flags 库会自动检测到 Version 字段的 long:"version" tag， 3.库会自动调用 optionsData.Version() 函数
	optionsData.Version = func() {
		err := printVersions()
		if err != nil {
			panic(&exitStatus{1})
		}
		panic(&exitStatus{0})
	}
	
	// 设置命令行解析选项：允许使用双破折号 "--" 分隔选项和参数
	// flags.PassDoubleDash 表示当遇到单独的 "--" 时，停止解析选项
	// 后续参数将被视为位置参数（即使以 - 开头）
	flagopts := flags.Options(flags.PassDoubleDash)
	// logf("flagopts: %v", flagopts) //输出 4 表示 flags.PassDoubleDash 的枚举值
	// NewParser 参数说明：
	// 1. &optionsData: 包含命令选项定义的结构体（含 --version 回调）
	// 2. flagopts: 解析器选项组合（此处为二进制位掩码）
	parser := flags.NewParser(&optionsData, flagopts)
	// 配置解析器的元信息
	parser.ShortDescription = "Tool to interact with chisel"  // 设置简短描述（帮助信息首行）
	parser.LongDescription = longChiselDescription          // 设置详细描述（帮助内容主体）
	
	// 隐藏默认的"[OPTIONS]"占位符
	parser.Usage = ""  // 清空自动生成的用法提示
	
	// 查找并定制--version选项
	if version := parser.FindOptionByLongName("version"); version != nil {
		version.Description = "Print the version and exit"  // 覆盖默认描述
		version.Hidden = true  // 在帮助信息中隐藏该选项
	}
	
	// 手动添加--help选项（覆盖库的自动添加行为）
	err := addHelp(parser)  // 添加自定义帮助选项
	if err != nil {
		debugf("cannot add --help: %v", err)  // 记录调试信息（不影响主流程）
	}
	// logf("commands: %+v", commands)
	// Add all regular commands
	// commands来自在每个包的init函数中执行带入：Go中的init函数在包被导入时自动执行，且每个包的init函数按照导入顺序执行，最后是main包的init函数，然后是main函数
	for _, c := range commands {
		//
		obj := c.builder()
		//if x, ok := obj.(clientSetter); ok {
		//	x.setClient(cli)
		//}
		if x, ok := obj.(parserSetter); ok {
			x.setParser(parser)
		}

		cmd, err := parser.AddCommand(c.name, c.shortHelp, strings.TrimSpace(c.longHelp), obj)
		if err != nil {
			panicf("cannot add command %q: %v", c.name, err)
		}
		cmd.Hidden = c.hidden
		if c.alias != "" {
			cmd.Aliases = append(cmd.Aliases, c.alias)
		}

		opts := cmd.Options()
		if c.optDescs != nil && len(opts) != len(c.optDescs) {
			panicf("wrong number of option descriptions for %s: expected %d, got %d", c.name, len(opts), len(c.optDescs))
		}
		for _, opt := range opts {
			name := opt.LongName
			if name == "" {
				name = string(opt.ShortName)
			}
			desc, ok := c.optDescs[name]
			if !(c.optDescs == nil || ok) {
				panicf("%s missing description for %s", c.name, name)
			}
			lintDesc(c.name, name, desc, opt.Description)
			if desc != "" {
				opt.Description = desc
			}
		}

		args := cmd.Args()
		if c.argDescs != nil && len(args) != len(c.argDescs) {
			panicf("wrong number of argument descriptions for %s: expected %d, got %d", c.name, len(args), len(c.argDescs))
		}
		for i, arg := range args {
			name, desc := arg.Name, ""
			if c.argDescs != nil {
				name = c.argDescs[i].name
				desc = c.argDescs[i].desc
			}
			lintArg(c.name, name, desc, arg.Description)
			name = fixupArg(name)
			arg.Name = name
			arg.Description = desc
		}
		if c.extra != nil {
			c.extra(cmd)
		}
	}
	// Add the debug command
	debugCommand, err := parser.AddCommand("debug", shortDebugHelp, longDebugHelp, &cmdDebug{})
	debugCommand.Hidden = true
	if err != nil {
		panicf("cannot add command %q: %v", "debug", err)
	}
	// Add all the sub-commands of the debug command
	for _, c := range debugCommands {
		obj := c.builder()
		//if x, ok := obj.(clientSetter); ok {
		//	x.setClient(cli)
		//}
		cmd, err := debugCommand.AddCommand(c.name, c.shortHelp, strings.TrimSpace(c.longHelp), obj)
		if err != nil {
			panicf("cannot add debug command %q: %v", c.name, err)
		}
		cmd.Hidden = c.hidden
		opts := cmd.Options()
		if c.optDescs != nil && len(opts) != len(c.optDescs) {
			panicf("wrong number of option descriptions for %s: expected %d, got %d", c.name, len(opts), len(c.optDescs))
		}
		for _, opt := range opts {
			name := opt.LongName
			if name == "" {
				name = string(opt.ShortName)
			}
			desc, ok := c.optDescs[name]
			if !(c.optDescs == nil || ok) {
				panicf("%s missing description for %s", c.name, name)
			}
			lintDesc(c.name, name, desc, opt.Description)
			if desc != "" {
				opt.Description = desc
			}
		}

		args := cmd.Args()
		if c.argDescs != nil && len(args) != len(c.argDescs) {
			panicf("wrong number of argument descriptions for %s: expected %d, got %d", c.name, len(args), len(c.argDescs))
		}
		for i, arg := range args {
			name, desc := arg.Name, ""
			if c.argDescs != nil {
				name = c.argDescs[i].name
				desc = c.argDescs[i].desc
			}
			lintArg(c.name, name, desc, arg.Description)
			name = fixupArg(name)
			arg.Name = name
			arg.Description = desc
		}
	}
	return parser
}

var (
	isStdinTTY  = term.IsTerminal(0)
	isStdoutTTY = term.IsTerminal(1)
)

func main() {
	// defer是程序的最后一道异常处理屏障，与exitStatus结构体配合使用，实现了程序的优雅退出控制
	defer func() {
		// 使用recover()函数捕获程序运行时的panic异常
		// 当有panic发生时，v会获得非nil的错误值
		if v := recover(); v != nil {
			// 检查panic值是否是*exitStatus类型（自定义的退出状态类型）
			// 如果是则调用os.Exit()以指定的状态码退出程序
			if e, ok := v.(*exitStatus); ok {
				os.Exit(e.code)
			}
			// 对于非exitStatus类型的panic，重新抛出异常
			panic(v)
		}
	}()

	if err := run(); err != nil {
		fmt.Fprintf(Stderr, errorPrefix+"%v\n", err)
		os.Exit(1)
	}
}

// exitStatus can be used in panic(&exitStatus{code}) to cause Chisel's main
// function to exit with a given exit code, for the rare cases when you want
// to return an exit code other than 0 or 1, or when an error return is not
// possible.
// 1. panic(&exitStatus{255})会被defer捕获，转换为os.Exit(255)，程序以状态码255退出。
// 2. 其他panic（如panic("error")）会被重新抛出，导致程序崩溃，并打印堆栈跟踪
type exitStatus struct {
	code int
}

func (e *exitStatus) Error() string {
	return fmt.Sprintf("internal error: exitStatus{%d} being handled as normal error", e.code)
}

func run() error {
	archive.SetLogger(log.Default())
	rpm.SetLogger(log.Default())
	setup.SetLogger(log.Default())
	slicer.SetLogger(log.Default())
	SetLogger(log.Default())

	parser := Parser()
	xtra, err := parser.Parse()
	if err != nil {
		if e, ok := err.(*flags.Error); ok {
			switch e.Type {
			case flags.ErrCommandRequired:
				printShortHelp()
				return nil
			case flags.ErrHelp:
				parser.WriteHelp(Stdout)
				return nil
			case flags.ErrUnknownCommand:
				sub := os.Args[1]
				sug := "chisel help"
				if len(xtra) > 0 {
					sub = xtra[0]
					if x := parser.Command.Active; x != nil && x.Name != "help" {
						sug = "chisel help " + x.Name
					}
				}
				return fmt.Errorf("unknown command %q, see '%s'.", sub, sug)
			}
		}
		return err
	}

	return nil
}

var errorPrefix = "error: "
