package setup

import (
	"fmt"
	"os"
	"path/filepath"
	"regexp"
	"strings"

	"golang.org/x/crypto/openpgp/packet"

	"github.com/canonical/chisel/internal/strdist"
)

// Release is a collection of package slices targeting a particular
// distribution version.
type Release struct {
	Path     string
	Packages map[string]*Package
	Archives map[string]*Archive
}

// Archive is the location from which binary packages are obtained.
type Archive struct {
	Name       string
	Version    string
	Suites     []string
	Components []string
	Priority   int
	Pro        string
	PubKeys    []*packet.PublicKey
}

// Package holds a collection of slices that represent parts of themselves.
type Package struct {
	Name    string
	Path    string
	Archive string
	Slices  map[string]*Slice
}

// Slice holds the details about a package slice.
type Slice struct {
	Package   string
	Name      string
	Essential []SliceKey
	Contents  map[string]PathInfo
	Scripts   SliceScripts
}

type SliceScripts struct {
	Mutate string
}

type PathKind string

const (
	DirPath      PathKind = "dir"
	CopyPath     PathKind = "copy"
	GlobPath     PathKind = "glob"
	TextPath     PathKind = "text"
	SymlinkPath  PathKind = "symlink"
	GeneratePath PathKind = "generate"

	// TODO Maybe in the future, for binary support.
	//Base64Path PathKind = "base64"
)

type PathUntil string

const (
	UntilNone   PathUntil = ""
	UntilMutate PathUntil = "mutate"
)

type GenerateKind string

const (
	GenerateNone     GenerateKind = ""
	GenerateManifest GenerateKind = "manifest"
)

type PathInfo struct {
	Kind PathKind
	Info string
	Mode uint

	Mutable  bool
	Until    PathUntil
	Arch     []string
	Generate GenerateKind
}

// SameContent returns whether the path has the same content properties as some
// other path. In other words, the resulting file/dir entry is the same. The
// Mutable flag must also match, as that's a common agreement that the actual
// content is not well defined upfront.
func (pi *PathInfo) SameContent(other *PathInfo) bool {
	return (pi.Kind == other.Kind &&
		pi.Info == other.Info &&
		pi.Mode == other.Mode &&
		pi.Mutable == other.Mutable &&
		pi.Generate == other.Generate)
}

type SliceKey struct {
	Package string
	Slice   string
}

func (s *Slice) String() string   { return s.Package + "_" + s.Name }
func (s SliceKey) String() string { return s.Package + "_" + s.Slice }

// Selection holds the required configuration to create a Build for a selection
// of slices from a Release. It's still an abstract proposal in the sense that
// the real information coming from pacakges is still unknown, so referenced
// paths could potentially be missing, for example.
// Selection 包含创建 Build 所需的配置，用于从 Release 中选择
// 切片。它仍然是一个抽象的提议，因为
// 来自包的真实信息仍然未知，因此引用的
// 路径可能会丢失。
type Selection struct {
	Release *Release
	Slices  []*Slice
}

func ReadRelease(dir string) (*Release, error) {
	logDir := dir
	logf("logDir是: %s", logDir)
	if strings.Contains(dir, "/.cache/") {
		logDir = filepath.Base(dir)
		logf("由于包含.cache, logDir改为: %s", logDir)
	}
	logf("Processing %s release...", logDir)

	release := &Release{
		Path:     dir,
		Packages: make(map[string]*Package),
	}

	logf("准备在 %s 中 readRelease...", dir)
	release, err := readRelease(dir)
	//debugutil.SaveToFile("readRelease", release)
	if err != nil {
		return nil, err
	}
	logf("已经获取release")

	logf("执行validate, 检查切片是否存在文件冲突...")
	err = release.validate()
	if err != nil {
		return nil, err
	}
	logf("完成冲突检查")
	return release, nil
}

func (r *Release) validate() error {
	keys := []SliceKey(nil)

	// Check for info conflicts and prepare for following checks. A conflict
	// means that two slices attempt to extract different files or directories
	// to the same location.
	// Conflict validation is done without downloading packages which means that
	// if we are extracting content from different packages to the same location
	// we cannot be sure that it will be the same. On the contrary, content
	// extracted from the same package will never conflict because it is
	// guaranteed to be the same.
	// The above also means that generated content (e.g. text files, directories
	// with make:true) will always conflict with extracted content, because we
	// cannot validate that they are the same without downloading the package.
	// 检查信息冲突并准备进行后续检查。冲突意味着两个切片尝试将不同的文件或目录提取到同一位置。
	// 冲突验证是在不下载软件包的情况下完成的，这意味着如果我们将不同软件包的内容提取到同一位置，
	// 我们不能确定它们是否相同。相反，从同一软件包中提取的内容永远不会发生冲突，因为它们保证是相同的。
	// 以上还意味着生成的内容（例如文本文件、带有 make:true 的目录）将始终与提取的内容冲突，
	// 因为我们无法在不下载软件包的情况下验证它们是否相同。
	paths := make(map[string]*Slice)
	globs := make(map[string]*Slice)
	for _, pkg := range r.Packages {
		for _, new := range pkg.Slices {
			// 遍历每个包的每个slice
			keys = append(keys, SliceKey{pkg.Name, new.Name})
			for newPath, newInfo := range new.Contents {
				if old, ok := paths[newPath]; ok {
					// 判断paths里是否含有newPath这个key值，有的话将value赋值给old
					oldInfo := old.Contents[newPath]
					// 这段代码的目的是检测两个信息结构体 newInfo 和 oldInfo 是否存在内容冲突，特别是在路径类型为 CopyPath 或 GlobPath 且包名不同的情况下。如果检测到冲突，则根据包名和名称的字典序进行排序，并返回一个错误，指出在特定路径上存在冲突。
					if !newInfo.SameContent(&oldInfo) || (newInfo.Kind == CopyPath || newInfo.Kind == GlobPath) && new.Package != old.Package {
						if old.Package > new.Package || old.Package == new.Package && old.Name > new.Name {
							old, new = new, old
						}
						return fmt.Errorf("slices %s and %s conflict on %s", old, new, newPath)
					}
					// Note: Because for conflict resolution we only check that
					// the created file would be the same and we know newInfo and
					// oldInfo produce the same one, we do not have to record
					// newInfo.
					// 注意：因为为了解决冲突，我们只检查创建的文件是否相同，
					// 并且我们知道 newInfo 和 oldInfo 产生相同的文件，所以我们不必记录 newInfo。
				} else {
					paths[newPath] = new
					// GlobPath 是用于通配符匹配的路径类型。
					// GeneratePath 是通过代码生成机制得到的路径类型。
					if newInfo.Kind == GeneratePath || newInfo.Kind == GlobPath {
						globs[newPath] = new
					}
				}
			}
		}
	}

	// Check for glob and generate conflicts.
	for oldPath, old := range globs {
		oldInfo := old.Contents[oldPath]
		for newPath, new := range paths {
			if oldPath == newPath {
				// Identical paths have been filtered earlier. This must be the
				// exact same entry.
				// 相同的路径先前已被过滤。这必须是完全相同的条目。
				continue
			}
			newInfo := new.Contents[newPath]
			//如果旧路径是一个 GlobPath，并且新路径也是一个 GlobPath 或 CopyPath，则需要进一步检查它们的包名。
			//如果新路径和旧路径的包名相同，则认为它们不会产生冲突，因此跳过当前循环。
			//如果包名不同，则可能会产生冲突，需要进一步处理。
			if oldInfo.Kind == GlobPath && (newInfo.Kind == GlobPath || newInfo.Kind == CopyPath) {
				if new.Package == old.Package {
					continue
				}
			}
			if strdist.GlobPath(newPath, oldPath) {
				//strdist.GlobPath检查两个字符串 newPath 和 oldPath 是否匹配，支持 ?、* 和 ** 三种通配符。
				if (old.Package > new.Package) || (old.Package == new.Package && old.Name > new.Name) ||
					(old.Package == new.Package && old.Name == new.Name && oldPath > newPath) {
					old, new = new, old
					oldPath, newPath = newPath, oldPath
				}
				return fmt.Errorf("slices %s and %s conflict on %s and %s", old, new, oldPath, newPath)
			}
		}
	}

	// Check for cycles.
	logf("开始执行order, 检查是否有循环依赖")
	//这里的keys里的每一个Slicekey，实际上是每个Package的Name以及他的每个slice的name组合在一起
	_, err := order(r.Packages, keys)
	if err != nil {
		return err
	}
	logf("完成执行order, 无循环依赖")

	// Check for archive priority conflicts.
	priorities := make(map[int]*Archive)
	for _, archive := range r.Archives {
		if old, ok := priorities[archive.Priority]; ok {
			if old.Name > archive.Name {
				archive, old = old, archive
			}
			return fmt.Errorf("chisel.yaml: archives %q and %q have the same priority value of %d", old.Name, archive.Name, archive.Priority)
		}
		priorities[archive.Priority] = archive
	}

	// Check that archives pinned in packages are defined.
	for _, pkg := range r.Packages {
		if pkg.Archive == "" {
			continue
		}
		if _, ok := r.Archives[pkg.Archive]; !ok {
			return fmt.Errorf("%s: package refers to undefined archive %q", pkg.Path, pkg.Archive)
		}
	}

	return nil
}

func order(pkgs map[string]*Package, keys []SliceKey) ([]SliceKey, error) {

	// Preprocess the list to improve error messages.
	for _, key := range keys {
		if pkg, ok := pkgs[key.Package]; !ok {
			return nil, fmt.Errorf("slices of package %q not found", key.Package)
		} else if _, ok := pkg.Slices[key.Slice]; !ok {
			return nil, fmt.Errorf("slice %s not found", key)
		}
	}

	// Collect all relevant package slices.
	//successors其键（key）是字符串类型，值（value）是字符串切片（slice）类型。
	successors := map[string][]string{}
	//pending 被赋值为一个新的切片，它包含了 keys 中的所有元素，但是是一个新的切片实例，而不是对 keys 的引用。
	pending := append([]SliceKey(nil), keys...)

	//seen用作标记SliceKey是否出现过一次
	seen := make(map[SliceKey]bool)
	for i := 0; i < len(pending); i++ {
		key := pending[i]
		if seen[key] {
			continue
		}
		seen[key] = true
		pkg := pkgs[key.Package]       //pkg是Package类型的
		slice := pkg.Slices[key.Slice] //slice是Slice类型的
		//fqslice是pkg和slice的拼接
		fqslice := slice.String()
		//当你从一个映射（map）中获取一个不存在的键（key）时，会返回该类型的零值。对于切片（slice）类型，零值是 nil
		predecessors := successors[fqslice]
		for _, req := range slice.Essential {
			// req是SliceKey类型的
			fqreq := req.String()
			// fqreq是req.Package和reqSlice的拼接
			if reqpkg, ok := pkgs[req.Package]; !ok || reqpkg.Slices[req.Slice] == nil {
				return nil, fmt.Errorf("%s requires %s, but slice is missing", fqslice, fqreq)
			}
			predecessors = append(predecessors, fqreq)
		}
		successors[fqslice] = predecessors
		//所以整了半天, successor的key是Slice类型的结构体执行string()的结果，其是pkg名字和slice类别（name）的拼接（因为一个pkg可以有不同类型的slice：lib类，copyright类等）
		//successors的value是SliceKey类型的结构体执行string()的结果， SliceKey是pkg名字和slice类别（name）的拼接，表明这个slice是依赖哪些其他slice的集合
		pending = append(pending, slice.Essential...)
		//append 添加了新元素，len(pending) 会增加，循环可能会继续执行，直到处理完所有新添加的元素。
	}
	//所以successors包含了所有依赖关系，即某个切片依赖哪些其他slice，组成了一个所谓的图，然后通过tarjan算法计算出拓扑排序，如果拓扑排序的结果中存在环，那么就说明存在循环依赖。
	// Sort them up.
	var order []SliceKey
	for _, names := range tarjanSort(successors) {
		// 遍历每个强连通分量
		if len(names) > 1 {
			//说明当前强连通分量包含多个节点，这意味着存在一个循环依赖。
			return nil, fmt.Errorf("essential loop detected: %s", strings.Join(names, ", "))
		}
		name := names[0]
		dot := strings.IndexByte(name, '_') //找到 name 中第一个下划线 _ 的位置。
		// name[:dot] 是 name 中下划线之前的部分，表示 Package 名称。
		// name[dot+1:] 是 name 中下划线之后的部分，表示 Slice 名称。
		order = append(order, SliceKey{name[:dot], name[dot+1:]})
	}

	return order, nil
}

// fnameExp matches the slice definition file basename.
var fnameExp = regexp.MustCompile(`^([a-z0-9](?:-?[.a-z0-9+]){1,})\.yaml$`)

// snameExp matches only the slice name, without the leading package name.
var snameExp = regexp.MustCompile(`^([a-z](?:-?[a-z0-9]){2,})$`)

// knameExp matches the slice full name in pkg_slice format.
var knameExp = regexp.MustCompile(`^([a-z0-9](?:-?[.a-z0-9+]){1,})_([a-z](?:-?[a-z0-9]){2,})$`)

func ParseSliceKey(sliceKey string) (SliceKey, error) {
	match := knameExp.FindStringSubmatch(sliceKey)
	if match == nil {
		return SliceKey{}, fmt.Errorf("invalid slice reference: %q", sliceKey)
	}
	return SliceKey{match[1], match[2]}, nil
}

func readRelease(baseDir string) (*Release, error) {
	baseDir = filepath.Clean(baseDir)
	logf("先清理 %s", baseDir)
	filePath := filepath.Join(baseDir, "chisel.yaml")
	logf("filePath: %s", filePath)
	data, err := os.ReadFile(filePath)
	if err != nil {
		return nil, fmt.Errorf("cannot read release definition: %s", err)
	}
	logf("获取chisel.yaml的内容, 进入parseRelease")
	release, err := parseRelease(baseDir, filePath, data)
	if err != nil {
		return nil, err
	}
	logf("进入readSlices, 获取slices目录下的内容")
	err = readSlices(release, baseDir, filepath.Join(baseDir, "slices"))
	if err != nil {
		return nil, err
	}
	return release, err
}

func readSlices(release *Release, baseDir, dirName string) error {
	entries, err := os.ReadDir(dirName)
	if err != nil {
		return fmt.Errorf("cannot read %s%c directory", stripBase(baseDir, dirName), filepath.Separator)
	}

	// 这段代码逻辑是遍历解析slice文件夹下的每个package.yaml文件，
	for _, entry := range entries {
		if entry.IsDir() {
			err := readSlices(release, baseDir, filepath.Join(dirName, entry.Name()))
			if err != nil {
				return err
			}
			continue
		}
		if entry.IsDir() || !strings.HasSuffix(entry.Name(), ".yaml") {
			continue
		}
		//正则表达式匹配^([a-z0-9](?:-?[.a-z0-9+]){1,})\.yaml$
		match := fnameExp.FindStringSubmatch(entry.Name())
		if match == nil {
			return fmt.Errorf("invalid slice definition filename: %q", entry.Name())
		}

		pkgName := match[1]
		pkgPath := filepath.Join(dirName, entry.Name())
		if pkg, ok := release.Packages[pkgName]; ok {
			return fmt.Errorf("package %q slices defined more than once: %s and %s\")", pkgName, pkg.Path, pkgPath)
		}
		data, err := os.ReadFile(pkgPath)
		if err != nil {
			// Errors from package os generally include the path.
			return fmt.Errorf("cannot read slice definition file: %v", err)
		}

		pkg, err := parsePackage(baseDir, pkgName, stripBase(baseDir, pkgPath), data)
		if err != nil {
			return err
		}

		release.Packages[pkg.Name] = pkg
	}
	return nil
}

func stripBase(baseDir, path string) string {
	// Paths must be clean for this to work correctly.
	return strings.TrimPrefix(path, baseDir+string(filepath.Separator))
}

func Select(release *Release, slices []SliceKey) (*Selection, error) {
	logf("Selecting slices...")

	selection := &Selection{
		Release: release,
	}
	logf("执行order获取依赖 %v 的所有切片, 返回SliceKey切片集合", slices)
	//debugutil.SaveToFile("release.Packages", release.Packages)
	sorted, err := order(release.Packages, slices)
	logf("挑选出来的切片： %v", sorted)
	// sorted是SliceKey结构体
	if err != nil {
		return nil, err
	}
	selection.Slices = make([]*Slice, len(sorted))

	for i, key := range sorted {
		selection.Slices[i] = release.Packages[key.Package].Slices[key.Slice]
	}
	logf("检查挑选出来的切片是否存在路径冲突")
	paths := make(map[string]*Slice)
	for _, new := range selection.Slices {
		for newPath, newInfo := range new.Contents {
			if old, ok := paths[newPath]; ok {
				oldInfo := old.Contents[newPath]
				if !newInfo.SameContent(&oldInfo) || (newInfo.Kind == CopyPath || newInfo.Kind == GlobPath) && new.Package != old.Package {
					if old.Package > new.Package || old.Package == new.Package && old.Name > new.Name {
						old, new = new, old
					}
					return nil, fmt.Errorf("slices %s and %s conflict on %s", old, new, newPath)
				}
			} else {
				paths[newPath] = new
			}
			// An invalid "generate" value should only throw an error if that
			// particular slice is selected. Hence, the check is here.
			switch newInfo.Generate {
			case GenerateNone, GenerateManifest:
			default:
				return nil, fmt.Errorf("slice %s has invalid 'generate' for path %s: %q",
					new, newPath, newInfo.Generate)
			}
		}
	}

	return selection, nil
}
